@@ -2,7 +2,7 @@
 /*
  * Ruby Cairo Binding
  *
- * Copyright 2005-2017 Kouhei Sutou <kou@cozmixng.org>
+ * Copyright 2005-2018 Kouhei Sutou <kou@cozmixng.org>
  * Copyright 2014 Patrick Hanevold <patrick.hanevold@gmail.com>
  * Copyright 2005 Øyvind Kolås <pippin@freedesktop.org>
  * Copyright 2004-2005 MenTaLguY <mental@rydia.com>
@@ -90,6 +90,7 @@ VALUE rb_cCairo_SkiaSurface = Qnil;
 VALUE rb_cCairo_SubSurface = Qnil;
 VALUE rb_cCairo_CoglSurface = Qnil;
 
+static ID cr_id_new;
 static ID cr_id_parse;
 static ID cr_id_size;
 static ID cr_id_set_unit;
@@ -217,6 +218,67 @@ cr_surface_get_klass (cairo_surface_t *surface)
   return klass;
 }
 
+static void
+rb_cairo_surface_adjust_memory_usage (cairo_surface_t *surface,
+                                      cairo_bool_t new)
+{
+#ifdef HAVE_RB_GC_ADJUST_MEMORY_USAGE
+  if (cairo_surface_get_type (surface) == CAIRO_SURFACE_TYPE_IMAGE)
+    {
+      ssize_t memory_usage_diff;
+      memory_usage_diff =
+        cairo_image_surface_get_stride (surface) *
+        cairo_image_surface_get_height (surface);
+      if (!new)
+        memory_usage_diff = -memory_usage_diff;
+      rb_gc_adjust_memory_usage (memory_usage_diff);
+    }
+#endif
+}
+
+static void
+cr_surface_destroy_raw (cairo_surface_t *surface)
+{
+  rb_cairo_surface_adjust_memory_usage (surface, CR_FALSE);
+  cairo_surface_destroy (surface);
+}
+
+static VALUE
+cr_surface_destroy (VALUE self)
+{
+  cairo_surface_t *surface;
+
+  surface = _SELF;
+  cr_surface_destroy_raw (surface);
+  DATA_PTR (self) = NULL;
+
+  return self;
+}
+
+static VALUE
+cr_surface_destroy_with_destroy_check (VALUE self)
+{
+  if (_SELF)
+    cr_surface_destroy (self);
+  return Qnil;
+}
+
+static VALUE
+cr_surface_create (int argc, VALUE *argv, VALUE klass)
+{
+  VALUE rb_surface;
+  rb_surface = rb_funcallv (klass, cr_id_new, argc, argv);
+  if (rb_block_given_p ())
+    {
+      return rb_ensure (rb_cairo__surface_yield_and_finish, rb_surface,
+                        cr_surface_destroy_with_destroy_check, rb_surface);
+    }
+  else
+    {
+      return rb_surface;
+    }
+}
+
 static VALUE
 cr_surface_image_supported_p (VALUE klass)
 {
@@ -349,24 +411,6 @@ cr_surface_xml_supported_p (VALUE klass)
 #endif
 }
 
-static void
-rb_cairo_surface_adjust_memory_usage (cairo_surface_t *surface,
-                                      cairo_bool_t new)
-{
-#ifdef HAVE_RB_GC_ADJUST_MEMORY_USAGE
-  if (cairo_surface_get_type (surface) == CAIRO_SURFACE_TYPE_IMAGE)
-    {
-      ssize_t memory_usage_diff;
-      memory_usage_diff =
-        cairo_image_surface_get_stride (surface) *
-        cairo_image_surface_get_height (surface);
-      if (!new)
-        memory_usage_diff = -memory_usage_diff;
-      rb_gc_adjust_memory_usage (memory_usage_diff);
-    }
-#endif
-}
-
 /* constructor/de-constructor */
 cairo_surface_t *
 rb_cairo_surface_from_ruby_object (VALUE obj)
@@ -402,9 +446,7 @@ cr_surface_free (void *ptr)
   if (!surface)
     return;
 
-  rb_cairo_surface_adjust_memory_usage (surface, CR_FALSE);
-
-  cairo_surface_destroy (surface);
+  cr_surface_destroy_raw (surface);
 }
 
 VALUE
@@ -453,18 +495,6 @@ cr_surface_initialize (int argc, VALUE *argv, VALUE self)
 }
 
 /* Surface manipulation */
-static VALUE
-cr_surface_destroy (VALUE self)
-{
-  cairo_surface_t *surface;
-
-  surface = _SELF;
-  cairo_surface_destroy (surface);
-  DATA_PTR (self) = NULL;
-
-  return self;
-}
-
 static VALUE
 cr_surface_get_reference_count (VALUE self)
 {
@@ -497,20 +527,25 @@ cr_surface_finish (VALUE self)
   return self;
 }
 
-void
+VALUE
 rb_cairo__surface_yield_and_finish (VALUE self)
 {
+  VALUE rb_result;
   cairo_surface_t *surface;
 
-  rb_yield (self);
+  rb_result = rb_yield (self);
 
   surface = _SELF;
+  if (!surface)
+    return rb_result;
   if (cairo_surface_status (surface) != CAIRO_STATUS_SUCCESS)
-    return;
+    return rb_result;
   if (cairo_surface_get_user_data (surface, &cr_finished_key))
-    return;
+    return rb_result;
 
   cr_surface_finish (self);
+
+  return rb_result;
 }
 
 static VALUE
@@ -632,14 +667,6 @@ cr_surface_unmap_image (VALUE self, VALUE rb_mapped_image)
 #endif
 
 #if CAIRO_CHECK_VERSION(1, 10, 0)
-static VALUE
-cr_surface_destroy_with_destroy_check (VALUE self)
-{
-  if (DATA_PTR (self))
-    cr_surface_destroy (self);
-  return Qnil;
-}
-
 static VALUE
 cr_surface_create_sub_rectangle_surface (VALUE self, VALUE x, VALUE y,
                                          VALUE width, VALUE height)
@@ -782,11 +809,22 @@ cr_surface_supported_mime_type_p (VALUE self, VALUE rb_mime_type)
 static VALUE
 cr_surface_get_font_options (VALUE self)
 {
-  cairo_font_options_t *options = cairo_font_options_create();
+  cairo_font_options_t *options;
+  VALUE rb_options;
+
+  options = cairo_font_options_create ();
+  rb_cairo_check_status (cairo_font_options_status (options));
+
+  /* TODO: Use rb_ensure() */
+  rb_options = CRFONTOPTIONS2RVAL (options);
+  cairo_font_options_destroy (options);
+
+  options = RVAL2CRFONTOPTIONS (rb_options);
   cairo_surface_get_font_options (_SELF, options);
   rb_cairo_surface_check_status (_SELF);
   rb_cairo_check_status (cairo_font_options_status (options));
-  return CRFONTOPTIONS2RVAL (options);
+
+  return rb_options;
 }
 
 static VALUE
@@ -1443,7 +1481,7 @@ cr_win32_surface_initialize (int argc, VALUE *argv, VALUE self)
         format = arg2;
         width = arg3;
         height = arg4;
-        win32_hdc = NIL_P (hdc) ? NULL : (HDC) NUM2UINT (hdc);
+        win32_hdc = NIL_P (hdc) ? NULL : (HDC) (UINT_PTR) NUM2UINT (hdc);
         surface = cairo_win32_surface_create_with_ddb (win32_hdc,
                                                        RVAL2CRFORMAT (format),
                                                        NUM2INT (width),
@@ -1504,6 +1542,27 @@ cr_svg_surface_restrict_to_version (VALUE self, VALUE version)
   rb_cairo_surface_check_status (_SELF);
   return Qnil;
 }
+
+#  if CAIRO_CHECK_VERSION(1, 15, 10)
+static VALUE
+cr_svg_surface_get_document_unit (VALUE self)
+{
+  cairo_svg_unit_t unit;
+  unit = cairo_svg_surface_get_document_unit (_SELF);
+  rb_cairo_surface_check_status (_SELF);
+  return INT2NUM (unit);
+}
+
+static VALUE
+cr_svg_surface_set_document_unit (VALUE self, VALUE rb_unit)
+{
+  cairo_svg_unit_t unit;
+  unit = RVAL2CRSVGUNIT (rb_unit);
+  cairo_svg_surface_set_document_unit (_SELF, unit);
+  rb_cairo_surface_check_status (_SELF);
+  return Qnil;
+}
+#  endif
 #endif
 
 #ifdef RB_CAIRO_HAS_WIN32_PRINTING_SURFACE
@@ -1946,6 +2005,7 @@ cr_finish_all_guarded_surfaces_at_end (VALUE data)
 void
 Init_cairo_surface (void)
 {
+  cr_id_new = rb_intern ("new");
   cr_id_parse = rb_intern ("parse");
   cr_id_size = rb_intern ("size");
   cr_id_set_unit = rb_intern ("unit=");
@@ -1961,6 +2021,8 @@ Init_cairo_surface (void)
   rb_cairo__initialize_gc_guard_holder_class (rb_cCairo_Surface);
   rb_set_end_proc(cr_finish_all_guarded_surfaces_at_end, Qnil);
 
+  rb_define_singleton_method (rb_cCairo_Surface, "create",
+                              cr_surface_create, -1);
   rb_define_singleton_method (rb_cCairo_Surface, "image_supported?",
                               cr_surface_image_supported_p, 0);
   rb_define_singleton_method (rb_cCairo_Surface, "pdf_supported?",
@@ -2180,6 +2242,12 @@ Init_cairo_surface (void)
                     cr_svg_surface_initialize, -1);
   rb_define_method (rb_cCairo_SVGSurface, "restrict_to_version",
                     cr_svg_surface_restrict_to_version, 1);
+#  if CAIRO_CHECK_VERSION(1, 15, 10)
+  rb_define_method (rb_cCairo_SVGSurface, "document_unit",
+                    cr_svg_surface_get_document_unit, 0);
+  rb_define_method (rb_cCairo_SVGSurface, "set_document_unit",
+                    cr_svg_surface_set_document_unit, 1);
+#  endif
 
   RB_CAIRO_DEF_SETTERS (rb_cCairo_SVGSurface);
 #endif
